Builder = Entity:extend()

--Patch job - FIX PLEASE
-- NO
gmouse = mouse
local mouse = {}




local builderMenu

mouse.primary = false
mouse.secondary = false

function Builder:new(x, y)
  Builder.super.new(self, x, y)
  self.type = "builder"
  self.id = nextEntityID
  nextEntityID = nextEntityID + 1

  self.selectedObject = nil
  self.selectedVertex = nil -- key of the selected polygon
  self.storedAttributes = {}
  self.showBuilderMenu = false

  self.mode = "polygon"


  builderMenu = require "builderMenu"
  builderMenu:populate(self)

end




function Builder:setMode(mode, attributes)
  self.mode = mode
  self.storedAttributes =  attributes or {}
  if self.activePolygon then
    self.activePolygon:delete()
  end
end


function love.wheelmoved(x, y)
  if builderMode and not builder.showBuilderMenu then
    camera.zoom = camera.zoom*1.07^y
  end
end


function Builder:showMenu()
  self.showBuilderMenu = true
end  

function Builder:hideMenu()
  self.showBuilderMenu = false
end  

function Builder:toggleMenu()
  self.showBuilderMenu = not self.showBuilderMenu
end


function Builder:isModeSelectable()
  return self.mode == "polygon" or self:isModeProp()
end

function Builder:isModeProp()
  return 
  self.mode == "prop" or 
  self.mode == "button" or 
  self.mode == "talkative" or 
  self.mode == "logicProp" or 
  self.mode == "door" or 
  self.mode == "slidingDoor" or 
  self.mode == "brazier" or 
  self.mode == "rainEmitter" or 
  self.mode == "light" or 
  self.mode == "directionStone"
end


function Builder:getPropImage()
  
  local propTexture = self.storedAttributes.textureName
  if self.mode == "door" then
    return texture.doors.wooden.closed
  elseif self.mode == "slidingDoor" then
    return texture.doors.wooden.closed
  elseif self.mode == "rainEmitter" then
    return texture.icon_rain
  elseif self.mode == "directionStone" then
    return texture.direction_stone
  else
    if texture[propTexture].main then
      return texture[propTexture].main
    else
      return texture[propTexture]
    end
  end
end



--All level builder code should be contained within this function

function Builder:update(dt)
  
  if self.showBuilderMenu then
    builderMenu:update()
  else
    local moveSpeed = 800
    
    
    if love.keyboard.isDown("r") then
      player1.body:setPosition(gmouse.worldX, gmouse.worldY)
    end
    
    
    if love.keyboard.isScancodeDown("lshift") then
      moveSpeed = 1600
    end
    
    if love.keyboard.isScancodeDown("right","d") then
      self.x = self.x + moveSpeed*dt/camera.zoom
    end
    if love.keyboard.isScancodeDown("left", "a") then
      self.x = self.x - moveSpeed*dt/camera.zoom
    end
    if love.keyboard.isScancodeDown("up", "w") then
      self.y = self.y - moveSpeed*dt/camera.zoom
    end
    if love.keyboard.isScancodeDown("down", "s") then
      self.y = self.y + moveSpeed*dt/camera.zoom
    end



    if self.mode == "polygon" then
      if not (self.selectedObject and self.selectedObject.type ~= "polygon") then

        if clickedThisFrame.primary then 
          if self.activePolygon then
            self.lastPoint = self.activePolygon:addPoint(self.activePolygon:treatCoordinates(gmouse.worldX, gmouse.worldY))
          else
            self.lastPoint = {}
            local transformedX, transformedY = lockToPixels(gmouse.worldX, gmouse.worldY)
            
            self.activePolygon = LevelPolygon(transformedX, transformedY, nil, copyTable(self.storedAttributes))
            table.insert(currentLevel.polygons, self.activePolygon)
            self.lastPoint = Vec2(gmouse.worldX, gmouse.worldY)
          end
        else
        
          if (self.lastPoint ~= {} and self.lastPoint and love.mouse.isDown(1)) then
            local distance = 30
            if vector.getMagnitudeSqrd(Vec2(gmouse.worldX, gmouse.worldY) - self.lastPoint) > (distance/camera.zoom)^2 and self.activePolygon then
              self.lastPoint = self.activePolygon:addPoint(self.activePolygon:treatCoordinates(gmouse.worldX, gmouse.worldY))
            end
          end
        end
      end
      
      
      if clickedThisFrame.secondary then
        if self.activePolygon then
          if self.activePolygon.numberOfPoints > 1 then
            self.activePolygon:removePoint()
          else
            self.activePolygon:delete()
          end
        end
      end
      
      
      if love.keyboard.isScancodeDown("return") then
        if self.activePolygon or (self.selectedObject and self.selectedObject.type == "levelPolygon") and not love.mouse.isDown(1) then
          if self.activePolygon and self.activePolygon.numberOfPoints > 2 then
            self.activePolygon:finalise()
            self.activePolygon = false
          elseif self.selectedObject then
            self.selectedObject:remake()
          end
          --currentLevel:saveLevel("autosaves/autosave "..os.date('%Y-%m-%d %H-%M-%S'))
        end
      end
      
      if not love.mouse.isDown(1) then
        self.selectedVertex = nil
      end
    elseif self:isModeProp() or self.mode == "flow" then
      if not self.selectedObject then
        
        local propTexture = self.storedAttributes.textureName
        local propImage = self:getPropImage()
        
        
        local translatedX, translatedY = nil, nil
        if self:isModeProp() or self.mode == "flow" then
          translatedX, translatedY = lockToPixels(gmouse.worldX, gmouse.worldY, "small", 
            propImage:getWidth()%2 == 1,
            propImage:getHeight()%2 == 1)
        end
      
      
      
      
      if clickedThisFrame.primary then
        
        local attributes = copyTable(self.storedAttributes)
        
        if self.mode == "prop" then
          Prop(translatedX, translatedY, attributes)
        elseif self.mode == "logicProp" then
          LogicProp(translatedX, translatedY, attributes)
        elseif self.mode == "talkative" then
          Talkative(translatedX, translatedY, attributes)
        elseif self.mode == "door" then
          Door(translatedX, translatedY, attributes)
        elseif self.mode == "slidingDoor" then
          SlidingDoor(translatedX, translatedY, attributes)
        elseif self.mode == "brazier" then
          Brazier(translatedX, translatedY, attributes)
        elseif self.mode == "rainEmitter" then
          RainEmitter(translatedX, translatedY, attributes)
        elseif self.mode == "light" then
          LightEmitter(translatedX, translatedY, attributes)
        elseif self.mode == "flow" then
          Flow(translatedX, translatedY, attributes)
        elseif self.mode == "directionStone" then
          DirectionStone(translatedX, translatedY, attributes)
        end
      end
    end
  elseif self.mode == "grass" then
    local translatedX, translatedY = lockToPixels(gmouse.worldX, gmouse.worldY, "small")
      
      
      if clickedThisFrame.primary then
        if self.activeGrass then
          self.activeGrass:addPoint(translatedX, translatedY)
        else
          self.activeGrass = Grass(translatedX, translatedY)
        end
      end
      
      
      
      
      if clickedThisFrame.secondary then
        if self.activeGrass then
          if #self.activeGrass.points > 1 then
            self.activeGrass:removePoint()
          else
            self.activeGrass.remove = true
            self.activeGrass = nil
            flushDrawTable()
          end
        end
      end
      
      
      
      if love.keyboard.isScancodeDown("return") then
        if self.activeGrass then
          self.activeGrass.finalised = true
          self.activeGrass = nil
          --currentLevel:saveLevel("autosaves/autosave "..os.date('%Y-%m-%d %H-%M-%S'))
        end
      end
      
      if self.activeGrass then
        self.activeGrass.nextPoint.x, self.activeGrass.nextPoint.y = self.activeGrass:getNextPoint(translatedX, translatedY)
      end

    elseif self.mode == "box" then
      local translatedX, translatedY = lockToPixels(gmouse.worldX, gmouse.worldY, "small")
      if clickedThisFrame.primary then
        PhysicsCube(translatedX, translatedY)
      end
    elseif self.mode == "fireflies" then
      local translatedX, translatedY = lockToPixels(gmouse.worldX, gmouse.worldY, "small")
      local distance = 50
      local drawDistance = 10
      self.mouseLastPos = self.mouseLastPos or gmouse
      if clickedThisFrame.primary then
        self.mouseLastPos = Vec2(gmouse.worldX, gmouse.worldY)
        Firefly(translatedX, translatedY)

      end
      local k = 0.5
      if love.mouse.isDown(1) and vector.getMagnitudeSqrd(vector.subtraction(Vec2(gmouse.worldX, gmouse.worldY), self.mouseLastPos)) >= drawDistance^2 then
        self.mouseLastPos = Vec2(gmouse.worldX, gmouse.worldY)
        Firefly(translatedX + love.math.random(-distance, distance)*k, translatedY + love.math.random(-distance, distance)*k)
      end
    end


    if (not self.activePolygon) and clickedThisFrame.secondary then
      if self:isModeSelectable() and (not self.selectedObject) then
        local mode = nil
        if self.mode == "polygon" then
          mode = "levelPolygon"
        elseif self:isModeProp() then
          mode = "prop"
        end        
        for i, object in ipairs(drawTable) do

          if object.type == mode or (mode == "prop" and object.isProp) then
            if object:checkSelect() then
              self.selectedObject = object
              object.selected = true
              break
            end
          end
        end 
      end
    end
  

    if love.keyboard.isScancodeDown("return") and not love.mouse.isDown(1) then
      if self.selectedObject then
        self.selectedObject.selected = false
        self.selectedObject = nil
--        currentLevel:saveLevel("autosaves/autosave "..os.date('%Y-%m-%d %H-%M-%S'))
      end
    end



    if (self:isModeSelectable()) and self.selectedObject then
      attributeEditor:update(self.selectedObject)
    end
  end
end


function Builder:draw()
  if not (screenshotMode or self.showBuilderMenu) then
    local offsetY = false
    local offsetX = false
    
    local propTexture = self.storedAttributes.textureName
    local propImage   = self:getPropImage()
    
    
    if self:isModeProp() or self.mode == "flow" then
      offsetX = propImage:getWidth()%2 == 1
      offsetY = propImage:getHeight()%2 == 1
    end
    
    
    
    local translatedX, translatedY = lockToPixels(gmouse.worldX, gmouse.worldY, "small", offsetX, offsetY)
    if self:isModeProp() and not self.selectedObject then
      if texture[propTexture].main then
        love.graphics.setColor(1,1,1,0.7)
        blue.graphics.drawCentredSet(texture[propTexture], translatedX, translatedY, 0, smallTextureScale, smallTextureScale)
      else
        love.graphics.setColor(1,1,1,0.7)
        blue.graphics.drawCentred(propImage, translatedX, translatedY, 0, smallTextureScale, smallTextureScale)
      end
      
    elseif self.mode == "grass" then
      if not self.activeGrass then
        love.graphics.setColor(1,1,1, 0.7)
        blue.graphics.drawCentred(texture.grass[1], translatedX, translatedY, 0, smallTextureScale, smallTextureScale)
      end
    elseif self.mode == "box" then
      love.graphics.setColor(1,1,1, 0.7)
      blue.graphics.drawCentred(texture.stoneBox, translatedX, translatedY, 0, smallTextureScale, smallTextureScale)
    elseif self.mode == "flow" then

      love.graphics.setColor(1,1,1, 0.7)
      blue.graphics.drawCentred(texture.flow, translatedX, translatedY, 0, smallTextureScale, smallTextureScale)

    end
    if self.selectedObject and self:isModeSelectable() then
      attributeEditor:draw()
    end
  end
  
  if self.showBuilderMenu then
    builderMenu:draw()
  end
end

